@@ -95,27 +95,25 @@ pub(crate) fn to_aexpr(expr: Expr, arena: &mut Arena<AExpr>) -> Node {
                 falsy: f,
             }
         }
-        Expr::Function {
+        Expr::AnonymousFunction {
             input,
             function,
             output_type,
             options,
-        } => AExpr::Function {
+        } => AExpr::AnonymousFunction {
             input: to_aexprs(input, arena),
             function,
             output_type,
             options,
         },
-        Expr::BinaryFunction {
-            input_a,
-            input_b,
+        Expr::Function {
+            input,
             function,
-            output_field,
-        } => AExpr::BinaryFunction {
-            input_a: to_aexpr(*input_a, arena),
-            input_b: to_aexpr(*input_b, arena),
+            options,
+        } => AExpr::Function {
+            input: to_aexprs(input, arena),
             function,
-            output_field,
+            options,
         },
         Expr::Shift { input, periods } => AExpr::Shift {
             input: to_aexpr(*input, arena),
@@ -138,13 +136,15 @@ pub(crate) fn to_aexpr(expr: Expr, arena: &mut Arena<AExpr>) -> Node {
             length,
         } => AExpr::Slice {
             input: to_aexpr(*input, arena),
-            offset,
-            length,
+            offset: to_aexpr(*offset, arena),
+            length: to_aexpr(*length, arena),
         },
         Expr::Wildcard => AExpr::Wildcard,
+        Expr::Count => AExpr::Count,
+        Expr::Nth(i) => AExpr::Nth(i),
         Expr::KeepName(_) => panic!("no keep_name expected at this point"),
         Expr::Exclude(_, _) => panic!("no exclude expected at this point"),
-        Expr::SufPreFix { .. } => panic!("no `suffix/prefix` expected at this point"),
+        Expr::RenameAlias { .. } => panic!("no `rename_alias` expected at this point"),
         Expr::Columns { .. } => panic!("no `columns` expected at this point"),
         Expr::DtypeColumn { .. } => panic!("no `dtype-columns` expected at this point"),
     };
@@ -155,17 +155,17 @@ pub(crate) fn to_alp(
     lp: LogicalPlan,
     expr_arena: &mut Arena<AExpr>,
     lp_arena: &mut Arena<ALogicalPlan>,
-) -> Node {
+) -> Result<Node> {
     let v = match lp {
         LogicalPlan::Union { inputs, options } => {
             let inputs = inputs
                 .into_iter()
                 .map(|lp| to_alp(lp, expr_arena, lp_arena))
-                .collect();
+                .collect::<Result<_>>()?;
             ALogicalPlan::Union { inputs, options }
         }
         LogicalPlan::Selection { input, predicate } => {
-            let i = to_alp(*input, expr_arena, lp_arena);
+            let i = to_alp(*input, expr_arena, lp_arena)?;
             let p = to_aexpr(predicate, expr_arena);
             ALogicalPlan::Selection {
                 input: i,
@@ -173,20 +173,18 @@ pub(crate) fn to_alp(
             }
         }
         LogicalPlan::Slice { input, offset, len } => {
-            let input = to_alp(*input, expr_arena, lp_arena);
+            let input = to_alp(*input, expr_arena, lp_arena)?;
             ALogicalPlan::Slice { input, offset, len }
         }
         LogicalPlan::Melt {
             input,
-            id_vars,
-            value_vars,
+            args,
             schema,
         } => {
-            let input = to_alp(*input, expr_arena, lp_arena);
+            let input = to_alp(*input, expr_arena, lp_arena)?;
             ALogicalPlan::Melt {
                 input,
-                id_vars,
-                value_vars,
+                args,
                 schema,
             }
         }
@@ -262,7 +260,7 @@ pub(crate) fn to_alp(
             schema,
         } => {
             let exp = expr.into_iter().map(|x| to_aexpr(x, expr_arena)).collect();
-            let i = to_alp(*input, expr_arena, lp_arena);
+            let i = to_alp(*input, expr_arena, lp_arena)?;
             ALogicalPlan::Projection {
                 expr: exp,
                 input: i,
@@ -275,7 +273,7 @@ pub(crate) fn to_alp(
             schema,
         } => {
             let exp = expr.into_iter().map(|x| to_aexpr(x, expr_arena)).collect();
-            let i = to_alp(*input, expr_arena, lp_arena);
+            let i = to_alp(*input, expr_arena, lp_arena)?;
             ALogicalPlan::LocalProjection {
                 expr: exp,
                 input: i,
@@ -285,9 +283,9 @@ pub(crate) fn to_alp(
         LogicalPlan::Sort {
             input,
             by_column,
-            reverse,
+            args,
         } => {
-            let input = to_alp(*input, expr_arena, lp_arena);
+            let input = to_alp(*input, expr_arena, lp_arena)?;
             let by_column = by_column
                 .into_iter()
                 .map(|x| to_aexpr(x, expr_arena))
@@ -295,15 +293,23 @@ pub(crate) fn to_alp(
             ALogicalPlan::Sort {
                 input,
                 by_column,
-                reverse,
+                args,
             }
         }
-        LogicalPlan::Explode { input, columns } => {
-            let input = to_alp(*input, expr_arena, lp_arena);
-            ALogicalPlan::Explode { input, columns }
+        LogicalPlan::Explode {
+            input,
+            columns,
+            schema,
+        } => {
+            let input = to_alp(*input, expr_arena, lp_arena)?;
+            ALogicalPlan::Explode {
+                input,
+                columns,
+                schema,
+            }
         }
         LogicalPlan::Cache { input } => {
-            let input = to_alp(*input, expr_arena, lp_arena);
+            let input = to_alp(*input, expr_arena, lp_arena)?;
             ALogicalPlan::Cache { input }
         }
         LogicalPlan::Aggregate {
@@ -315,7 +321,7 @@ pub(crate) fn to_alp(
             maintain_order,
             options,
         } => {
-            let i = to_alp(*input, expr_arena, lp_arena);
+            let i = to_alp(*input, expr_arena, lp_arena)?;
             let aggs_new = aggs.into_iter().map(|x| to_aexpr(x, expr_arena)).collect();
             let keys_new = keys
                 .iter()
@@ -340,8 +346,8 @@ pub(crate) fn to_alp(
             right_on,
             options,
         } => {
-            let i_l = to_alp(*input_left, expr_arena, lp_arena);
-            let i_r = to_alp(*input_right, expr_arena, lp_arena);
+            let input_left = to_alp(*input_left, expr_arena, lp_arena)?;
+            let input_right = to_alp(*input_right, expr_arena, lp_arena)?;
 
             let l_on = left_on
                 .into_iter()
@@ -353,8 +359,8 @@ pub(crate) fn to_alp(
                 .collect();
 
             ALogicalPlan::Join {
-                input_left: i_l,
-                input_right: i_r,
+                input_left,
+                input_right,
                 schema,
                 left_on: l_on,
                 right_on: r_on,
@@ -367,62 +373,58 @@ pub(crate) fn to_alp(
             schema,
         } => {
             let exp = exprs.into_iter().map(|x| to_aexpr(x, expr_arena)).collect();
-            let i = to_alp(*input, expr_arena, lp_arena);
+            let input = to_alp(*input, expr_arena, lp_arena)?;
             ALogicalPlan::HStack {
-                input: i,
+                input,
                 exprs: exp,
                 schema,
             }
         }
-        LogicalPlan::Distinct {
-            input,
-            maintain_order,
-            subset,
-        } => {
-            let i = to_alp(*input, expr_arena, lp_arena);
-            ALogicalPlan::Distinct {
-                input: i,
-                maintain_order,
-                subset,
-            }
+        LogicalPlan::Distinct { input, options } => {
+            let input = to_alp(*input, expr_arena, lp_arena)?;
+            ALogicalPlan::Distinct { input, options }
         }
         LogicalPlan::Udf {
             input,
             function,
-            projection_pd,
-            predicate_pd,
+            options,
             schema,
         } => {
-            let input = to_alp(*input, expr_arena, lp_arena);
+            let input = to_alp(*input, expr_arena, lp_arena)?;
             ALogicalPlan::Udf {
                 input,
                 function,
-                projection_pd,
-                predicate_pd,
+                options,
                 schema,
             }
         }
+        LogicalPlan::Error { err, .. } => {
+            // We just take the error. The LogicalPlan should not be used anymore once this
+            // is taken.
+            let mut err = err.lock();
+            return Err(err.take().unwrap());
+        }
     };
-    lp_arena.add(v)
+    Ok(lp_arena.add(v))
 }
 
-pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
+pub(crate) fn node_to_expr(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
     let expr = expr_arena.get(node).clone();
 
     match expr {
-        AExpr::Duplicated(node) => Expr::Duplicated(Box::new(node_to_exp(node, expr_arena))),
-        AExpr::IsUnique(node) => Expr::IsUnique(Box::new(node_to_exp(node, expr_arena))),
-        AExpr::Reverse(node) => Expr::Reverse(Box::new(node_to_exp(node, expr_arena))),
-        AExpr::Explode(node) => Expr::Explode(Box::new(node_to_exp(node, expr_arena))),
+        AExpr::Duplicated(node) => Expr::Duplicated(Box::new(node_to_expr(node, expr_arena))),
+        AExpr::IsUnique(node) => Expr::IsUnique(Box::new(node_to_expr(node, expr_arena))),
+        AExpr::Reverse(node) => Expr::Reverse(Box::new(node_to_expr(node, expr_arena))),
+        AExpr::Explode(node) => Expr::Explode(Box::new(node_to_expr(node, expr_arena))),
         AExpr::Alias(expr, name) => {
-            let exp = node_to_exp(expr, expr_arena);
+            let exp = node_to_expr(expr, expr_arena);
             Expr::Alias(Box::new(exp), name)
         }
         AExpr::Column(a) => Expr::Column(a),
         AExpr::Literal(s) => Expr::Literal(s),
         AExpr::BinaryExpr { left, op, right } => {
-            let l = node_to_exp(left, expr_arena);
-            let r = node_to_exp(right, expr_arena);
+            let l = node_to_expr(left, expr_arena);
+            let r = node_to_expr(right, expr_arena);
             Expr::BinaryExpr {
                 left: Box::new(l),
                 op,
@@ -430,15 +432,15 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             }
         }
         AExpr::Not(expr) => {
-            let exp = node_to_exp(expr, expr_arena);
+            let exp = node_to_expr(expr, expr_arena);
             Expr::Not(Box::new(exp))
         }
         AExpr::IsNotNull(expr) => {
-            let exp = node_to_exp(expr, expr_arena);
+            let exp = node_to_expr(expr, expr_arena);
             Expr::IsNotNull(Box::new(exp))
         }
         AExpr::IsNull(expr) => {
-            let exp = node_to_exp(expr, expr_arena);
+            let exp = node_to_expr(expr, expr_arena);
             Expr::IsNull(Box::new(exp))
         }
         AExpr::Cast {
@@ -446,7 +448,7 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             data_type,
             strict,
         } => {
-            let exp = node_to_exp(expr, expr_arena);
+            let exp = node_to_expr(expr, expr_arena);
             Expr::Cast {
                 expr: Box::new(exp),
                 data_type,
@@ -454,25 +456,25 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             }
         }
         AExpr::Sort { expr, options } => {
-            let exp = node_to_exp(expr, expr_arena);
+            let exp = node_to_expr(expr, expr_arena);
             Expr::Sort {
                 expr: Box::new(exp),
                 options,
             }
         }
         AExpr::Take { expr, idx } => {
-            let expr = node_to_exp(expr, expr_arena);
-            let idx = node_to_exp(idx, expr_arena);
+            let expr = node_to_expr(expr, expr_arena);
+            let idx = node_to_expr(idx, expr_arena);
             Expr::Take {
                 expr: Box::new(expr),
                 idx: Box::new(idx),
             }
         }
         AExpr::SortBy { expr, by, reverse } => {
-            let expr = node_to_exp(expr, expr_arena);
+            let expr = node_to_expr(expr, expr_arena);
             let by = by
                 .iter()
-                .map(|node| node_to_exp(*node, expr_arena))
+                .map(|node| node_to_expr(*node, expr_arena))
                 .collect();
             Expr::SortBy {
                 expr: Box::new(expr),
@@ -481,8 +483,8 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             }
         }
         AExpr::Filter { input, by } => {
-            let input = node_to_exp(input, expr_arena);
-            let by = node_to_exp(by, expr_arena);
+            let input = node_to_expr(input, expr_arena);
+            let by = node_to_expr(by, expr_arena);
             Expr::Filter {
                 input: Box::new(input),
                 by: Box::new(by),
@@ -490,36 +492,36 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
         }
         AExpr::Agg(agg) => match agg {
             AAggExpr::Min(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Min(Box::new(exp)).into()
             }
             AAggExpr::Max(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Max(Box::new(exp)).into()
             }
 
             AAggExpr::Median(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Median(Box::new(exp)).into()
             }
             AAggExpr::NUnique(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::NUnique(Box::new(exp)).into()
             }
             AAggExpr::First(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::First(Box::new(exp)).into()
             }
             AAggExpr::Last(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Last(Box::new(exp)).into()
             }
             AAggExpr::Mean(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Mean(Box::new(exp)).into()
             }
             AAggExpr::List(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::List(Box::new(exp)).into()
             }
             AAggExpr::Quantile {
@@ -527,7 +529,7 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
                 quantile,
                 interpol,
             } => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Quantile {
                     expr: Box::new(exp),
                     quantile,
@@ -536,28 +538,28 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
                 .into()
             }
             AAggExpr::Sum(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Sum(Box::new(exp)).into()
             }
             AAggExpr::Std(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Std(Box::new(exp)).into()
             }
             AAggExpr::Var(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Var(Box::new(exp)).into()
             }
             AAggExpr::AggGroups(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::AggGroups(Box::new(exp)).into()
             }
             AAggExpr::Count(expr) => {
-                let exp = node_to_exp(expr, expr_arena);
+                let exp = node_to_expr(expr, expr_arena);
                 AggExpr::Count(Box::new(exp)).into()
             }
         },
         AExpr::Shift { input, periods } => {
-            let e = node_to_exp(input, expr_arena);
+            let e = node_to_expr(input, expr_arena);
             Expr::Shift {
                 input: Box::new(e),
                 periods,
@@ -568,9 +570,9 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             truthy,
             falsy,
         } => {
-            let p = node_to_exp(predicate, expr_arena);
-            let t = node_to_exp(truthy, expr_arena);
-            let f = node_to_exp(falsy, expr_arena);
+            let p = node_to_expr(predicate, expr_arena);
+            let t = node_to_expr(truthy, expr_arena);
+            let f = node_to_expr(falsy, expr_arena);
 
             Expr::Ternary {
                 predicate: Box::new(p),
@@ -578,27 +580,25 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
                 falsy: Box::new(f),
             }
         }
-        AExpr::Function {
+        AExpr::AnonymousFunction {
             input,
             function,
             output_type,
             options,
-        } => Expr::Function {
+        } => Expr::AnonymousFunction {
             input: nodes_to_exprs(&input, expr_arena),
             function,
             output_type,
             options,
         },
-        AExpr::BinaryFunction {
-            input_a,
-            input_b,
+        AExpr::Function {
+            input,
             function,
-            output_field,
-        } => Expr::BinaryFunction {
-            input_a: Box::new(node_to_exp(input_a, expr_arena)),
-            input_b: Box::new(node_to_exp(input_b, expr_arena)),
+            options,
+        } => Expr::Function {
+            input: nodes_to_exprs(&input, expr_arena),
             function,
-            output_field,
+            options,
         },
         AExpr::Window {
             function,
@@ -606,9 +606,9 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             order_by,
             options,
         } => {
-            let function = Box::new(node_to_exp(function, expr_arena));
+            let function = Box::new(node_to_expr(function, expr_arena));
             let partition_by = nodes_to_exprs(&partition_by, expr_arena);
-            let order_by = order_by.map(|ob| Box::new(node_to_exp(ob, expr_arena)));
+            let order_by = order_by.map(|ob| Box::new(node_to_expr(ob, expr_arena)));
             Expr::Window {
                 function,
                 partition_by,
@@ -621,16 +621,18 @@ pub(crate) fn node_to_exp(node: Node, expr_arena: &Arena<AExpr>) -> Expr {
             offset,
             length,
         } => Expr::Slice {
-            input: Box::new(node_to_exp(input, expr_arena)),
-            offset,
-            length,
+            input: Box::new(node_to_expr(input, expr_arena)),
+            offset: Box::new(node_to_expr(offset, expr_arena)),
+            length: Box::new(node_to_expr(length, expr_arena)),
         },
+        AExpr::Count => Expr::Count,
+        AExpr::Nth(i) => Expr::Nth(i),
         AExpr::Wildcard => Expr::Wildcard,
     }
 }
 
 fn nodes_to_exprs(nodes: &[Node], expr_arena: &Arena<AExpr>) -> Vec<Expr> {
-    nodes.iter().map(|n| node_to_exp(*n, expr_arena)).collect()
+    nodes.iter().map(|n| node_to_expr(*n, expr_arena)).collect()
 }
 
 pub(crate) fn node_to_lp(
@@ -659,7 +661,7 @@ pub(crate) fn node_to_lp(
         }
         ALogicalPlan::Selection { input, predicate } => {
             let lp = node_to_lp(input, expr_arena, lp_arena);
-            let p = node_to_exp(predicate, expr_arena);
+            let p = node_to_expr(predicate, expr_arena);
             LogicalPlan::Selection {
                 input: Box::new(lp),
                 predicate: p,
@@ -677,7 +679,7 @@ pub(crate) fn node_to_lp(
             path,
             schema,
             options,
-            predicate: predicate.map(|n| node_to_exp(n, expr_arena)),
+            predicate: predicate.map(|n| node_to_expr(n, expr_arena)),
             aggregate: nodes_to_exprs(&aggregate, expr_arena),
         },
         #[cfg(feature = "ipc")]
@@ -691,7 +693,7 @@ pub(crate) fn node_to_lp(
         } => LogicalPlan::IpcScan {
             path,
             schema,
-            predicate: predicate.map(|n| node_to_exp(n, expr_arena)),
+            predicate: predicate.map(|n| node_to_expr(n, expr_arena)),
             aggregate: nodes_to_exprs(&aggregate, expr_arena),
             options,
         },
@@ -706,7 +708,7 @@ pub(crate) fn node_to_lp(
         } => LogicalPlan::ParquetScan {
             path,
             schema,
-            predicate: predicate.map(|n| node_to_exp(n, expr_arena)),
+            predicate: predicate.map(|n| node_to_expr(n, expr_arena)),
             aggregate: nodes_to_exprs(&aggregate, expr_arena),
             options,
         },
@@ -720,8 +722,8 @@ pub(crate) fn node_to_lp(
             schema,
             projection: projection
                 .as_ref()
-                .map(|nodes| nodes.iter().map(|n| node_to_exp(*n, expr_arena)).collect()),
-            selection: selection.map(|n| node_to_exp(n, expr_arena)),
+                .map(|nodes| nodes.iter().map(|n| node_to_expr(*n, expr_arena)).collect()),
+            selection: selection.map(|n| node_to_expr(n, expr_arena)),
         },
         ALogicalPlan::Projection {
             expr,
@@ -752,18 +754,26 @@ pub(crate) fn node_to_lp(
         ALogicalPlan::Sort {
             input,
             by_column,
-            reverse,
+            args,
         } => {
             let input = Box::new(node_to_lp(input, expr_arena, lp_arena));
             LogicalPlan::Sort {
                 input,
                 by_column: nodes_to_exprs(&by_column, expr_arena),
-                reverse,
+                args,
             }
         }
-        ALogicalPlan::Explode { input, columns } => {
+        ALogicalPlan::Explode {
+            input,
+            columns,
+            schema,
+        } => {
             let input = Box::new(node_to_lp(input, expr_arena, lp_arena));
-            LogicalPlan::Explode { input, columns }
+            LogicalPlan::Explode {
+                input,
+                columns,
+                schema,
+            }
         }
         ALogicalPlan::Cache { input } => {
             let input = Box::new(node_to_lp(input, expr_arena, lp_arena));
@@ -823,45 +833,36 @@ pub(crate) fn node_to_lp(
                 schema,
             }
         }
-        ALogicalPlan::Distinct {
-            input,
-            maintain_order,
-            subset,
-        } => {
+        ALogicalPlan::Distinct { input, options } => {
             let i = node_to_lp(input, expr_arena, lp_arena);
             LogicalPlan::Distinct {
                 input: Box::new(i),
-                maintain_order,
-                subset,
+                options,
             }
         }
         ALogicalPlan::Melt {
             input,
-            id_vars,
-            value_vars,
+            args,
             schema,
         } => {
             let input = node_to_lp(input, expr_arena, lp_arena);
             LogicalPlan::Melt {
                 input: Box::new(input),
-                id_vars,
-                value_vars,
+                args,
                 schema,
             }
         }
         ALogicalPlan::Udf {
             input,
             function,
-            predicate_pd,
-            projection_pd,
+            options,
             schema,
         } => {
             let input = Box::new(node_to_lp(input, expr_arena, lp_arena));
             LogicalPlan::Udf {
                 input,
                 function,
-                predicate_pd,
-                projection_pd,
+                options,
                 schema,
             }
         }
