@@ -3,7 +3,9 @@ use crate::physical_plan::PhysicalAggregation;
 use crate::prelude::*;
 use polars_arrow::export::arrow::{array::*, compute::concatenate::concatenate};
 use polars_arrow::prelude::QuantileInterpolOptions;
+use polars_arrow::utils::CustomIterTools;
 use polars_core::frame::groupby::{GroupByMethod, GroupsProxy};
+use polars_core::utils::NoNull;
 use polars_core::{prelude::*, POOL};
 use std::borrow::Cow;
 use std::sync::Arc;
@@ -89,9 +91,62 @@ impl PhysicalAggregation for AggregationExpr {
                 Ok(rename_option_series(agg_s, &keep_name))
             }
             GroupByMethod::Count => {
-                let mut ca = ac.groups.group_count();
-                ca.rename(&keep_name);
-                Ok(Some(ca.into_series()))
+                // a few fast paths that prevent materializing new groups
+                match ac.update_groups {
+                    UpdateGroups::WithSeriesLen => {
+                        let list = ac
+                            .series()
+                            .list()
+                            .expect("impl error, should be a list at this point");
+
+                        let mut s = match list.chunks().len() {
+                            1 => {
+                                let arr = list.downcast_iter().next().unwrap();
+                                let offsets = arr.offsets().as_slice();
+
+                                let mut previous = 0i64;
+                                let counts: NoNull<IdxCa> = offsets[1..]
+                                    .iter()
+                                    .map(|&o| {
+                                        let len = (o - previous) as IdxSize;
+                                        previous = o;
+                                        len
+                                    })
+                                    .collect_trusted();
+                                counts.into_inner()
+                            }
+                            _ => {
+                                let counts: NoNull<IdxCa> = list
+                                    .amortized_iter()
+                                    .map(|s| {
+                                        if let Some(s) = s {
+                                            s.as_ref().len() as IdxSize
+                                        } else {
+                                            1
+                                        }
+                                    })
+                                    .collect_trusted();
+                                counts.into_inner()
+                            }
+                        };
+                        s.rename(&keep_name);
+                        Ok(Some(s.into_series()))
+                    }
+                    UpdateGroups::WithGroupsLen => {
+                        // no need to update the groups
+                        // we can just get the attribute, because we only need the length,
+                        // not the correct order
+                        let mut ca = ac.groups.group_count();
+                        ca.rename(&keep_name);
+                        Ok(Some(ca.into_series()))
+                    }
+                    // materialize groups
+                    _ => {
+                        let mut ca = ac.groups().group_count();
+                        ca.rename(&keep_name);
+                        Ok(Some(ca.into_series()))
+                    }
+                }
             }
             GroupByMethod::First => {
                 let mut agg_s = ac.flat_naive().into_owned().agg_first(ac.groups());
@@ -251,15 +306,18 @@ impl PhysicalAggregation for AggQuantileExpr {
         groups: &GroupsProxy,
         state: &ExecutionState,
     ) -> Result<Option<Series>> {
-        let series = self.expr.evaluate(df, state)?;
-        let new_name = series.name().to_string();
-        let opt_agg = series.agg_quantile(groups, self.quantile, self.interpol);
+        let mut ac = self.expr.evaluate_on_groups(df, groups, state)?;
+        // don't change names by aggregations as is done in polars-core
+        let keep_name = ac.series().name().to_string();
 
+        let opt_agg =
+            ac.flat_naive()
+                .into_owned()
+                .agg_quantile(ac.groups(), self.quantile, self.interpol);
         let opt_agg = opt_agg.map(|mut agg| {
-            agg.rename(&new_name);
+            agg.rename(&keep_name);
             agg.into_series()
         });
-
         Ok(opt_agg)
     }
 }
@@ -307,11 +365,14 @@ impl PhysicalExpr for AggQuantileExpr {
     #[allow(clippy::ptr_arg)]
     fn evaluate_on_groups<'a>(
         &self,
-        _df: &DataFrame,
-        _groups: &'a GroupsProxy,
-        _state: &ExecutionState,
+        df: &DataFrame,
+        groups: &'a GroupsProxy,
+        state: &ExecutionState,
     ) -> Result<AggregationContext<'a>> {
-        unimplemented!()
+        let out = self.aggregate(df, groups, state)?.ok_or_else(|| {
+            PolarsError::ComputeError("Aggregation did not return a Series".into())
+        })?;
+        Ok(AggregationContext::new(out, Cow::Borrowed(groups), true))
     }
 
     fn to_field(&self, input_schema: &Schema) -> Result<Field> {
