@@ -1,6 +1,8 @@
 use super::*;
 use polars_arrow::prelude::QuantileInterpolOptions;
+use polars_core::frame::explode::MeltArgs;
 use polars_core::series::ops::NullBehavior;
+use polars_time::prelude::DateMethods;
 
 #[test]
 fn test_lazy_with_column() {
@@ -57,12 +59,17 @@ fn test_lazy_alias() {
 #[test]
 fn test_lazy_melt() {
     let df = get_df();
+
+    let args = MeltArgs {
+        id_vars: vec!["petal.width".to_string(), "petal.length".to_string()],
+        value_vars: vec!["sepal.length".to_string(), "sepal.width".to_string()],
+        variable_name: None,
+        value_name: None,
+    };
+
     let out = df
         .lazy()
-        .melt(
-            vec!["petal.width".to_string(), "petal.length".to_string()],
-            vec!["sepal.length".to_string(), "sepal.width".to_string()],
-        )
+        .melt(args)
         .filter(col("variable").eq(lit("sepal.length")))
         .select([col("variable"), col("petal.width"), col("value")])
         .collect()
@@ -627,16 +634,6 @@ fn test_lazy_reverse() {
         .frame_equal_missing(&df.reverse()))
 }
 
-#[test]
-fn test_lazy_predicate_pushdown_binary_expr() {
-    let df = load_df();
-    df.lazy()
-        .filter(col("a").eq(col("b")))
-        .select([col("c")])
-        .collect()
-        .unwrap();
-}
-
 #[test]
 fn test_lazy_update_column() {
     let df = load_df();
